@@ -1,9 +1,30 @@
 use data_encoding::HEXLOWER;
 use futures_util::StreamExt;
+#[cfg(feature = "alsa-backend")]
+use librespot::playback::mixer::alsamixer::AlsaMixer;
+use librespot::{
+    connect::{ConnectConfig, Spirc},
+    core::{
+        Session, SessionConfig, authentication::Credentials, cache::Cache, config::DeviceType,
+        version,
+    },
+    discovery::DnsSdServiceBuilder,
+    playback::{
+        audio_backend::{self, BACKENDS, SinkBuilder},
+        config::{
+            AudioFormat, Bitrate, NormalisationMethod, NormalisationType, PlayerConfig, VolumeCtrl,
+        },
+        dither,
+        mixer::{self, MixerConfig, MixerFn},
+        player::{Player, coefficient_to_duration, duration_to_coefficient},
+    },
+};
+use librespot_oauth::OAuthClientBuilder;
 use log::{debug, error, info, trace, warn};
 use sha1::{Digest, Sha1};
 use std::{
     env,
+    ffi::OsStr,
     fs::create_dir_all,
     ops::RangeInclusive,
     path::{Path, PathBuf},
@@ -14,31 +35,11 @@ use std::{
 };
 use sysinfo::{ProcessesToUpdate, System};
 use thiserror::Error;
+use tokio::sync::Semaphore;
 use url::Url;
 
-use librespot::{
-    connect::{config::ConnectConfig, spirc::Spirc},
-    core::{
-        authentication::Credentials, cache::Cache, config::DeviceType, version, Session,
-        SessionConfig,
-    },
-    discovery::DnsSdServiceBuilder,
-    playback::{
-        audio_backend::{self, SinkBuilder, BACKENDS},
-        config::{
-            AudioFormat, Bitrate, NormalisationMethod, NormalisationType, PlayerConfig, VolumeCtrl,
-        },
-        dither,
-        mixer::{self, MixerConfig, MixerFn},
-        player::{coefficient_to_duration, duration_to_coefficient, Player},
-    },
-};
-
-#[cfg(feature = "alsa-backend")]
-use librespot::playback::mixer::alsamixer::AlsaMixer;
-
 mod player_event_handler;
-use player_event_handler::{run_program_on_sink_events, EventHandler};
+use player_event_handler::{EventHandler, run_program_on_sink_events};
 
 fn device_id(name: &str) -> String {
     HEXLOWER.encode(&Sha1::digest(name.as_bytes()))
@@ -76,7 +77,9 @@ fn setup_logging(quiet: bool, verbose: bool) {
             builder.init();
 
             if verbose && quiet {
-                warn!("`--verbose` and `--quiet` are mutually exclusive. Logging can not be both verbose and quiet. Using verbose mode.");
+                warn!(
+                    "`--verbose` and `--quiet` are mutually exclusive. Logging can not be both verbose and quiet. Using verbose mode."
+                );
             }
         }
     }
@@ -220,7 +223,7 @@ struct Setup {
     zeroconf_backend: Option<DnsSdServiceBuilder>,
 }
 
-fn get_setup() -> Setup {
+async fn get_setup() -> Setup {
     const VALID_INITIAL_VOLUME_RANGE: RangeInclusive<u16> = 0..=100;
     const VALID_VOLUME_RANGE: RangeInclusive<f64> = 0.0..=100.0;
     const VALID_NORMALISATION_KNEE_RANGE: RangeInclusive<f64> = 0.0..=10.0;
@@ -276,6 +279,7 @@ fn get_setup() -> Setup {
     const VERSION: &str = "version";
     const VOLUME_CTRL: &str = "volume-ctrl";
     const VOLUME_RANGE: &str = "volume-range";
+    const VOLUME_STEPS: &str = "volume-steps";
     const ZEROCONF_PORT: &str = "zeroconf-port";
     const ZEROCONF_INTERFACE: &str = "zeroconf-interface";
     const ZEROCONF_BACKEND: &str = "zeroconf-backend";
@@ -291,6 +295,7 @@ fn get_setup() -> Setup {
     const DEVICE_SHORT: &str = "d";
     const VOLUME_CTRL_SHORT: &str = "E";
     const VOLUME_RANGE_SHORT: &str = "e";
+    const VOLUME_STEPS_SHORT: &str = ""; // no short flag
     const DEVICE_TYPE_SHORT: &str = "F";
     const FORMAT_SHORT: &str = "f";
     const DISABLE_AUDIO_CACHE_SHORT: &str = "G";
@@ -371,6 +376,8 @@ fn get_setup() -> Setup {
     #[cfg(not(feature = "alsa-backend"))]
     const VOLUME_RANGE_DESC: &str =
         "Range of the volume control (dB) from 0.0 to 100.0. Defaults to 60.0.";
+    const VOLUME_STEPS_DESC: &str =
+        "Number of incremental steps when responding to volume control updates. Defaults to 64.";
 
     let mut opts = getopts::Options::new();
     opts.optflag(
@@ -570,6 +577,12 @@ fn get_setup() -> Setup {
         VOLUME_RANGE_DESC,
         "RANGE",
     )
+    .optopt(
+        VOLUME_STEPS_SHORT,
+        VOLUME_STEPS,
+        VOLUME_STEPS_DESC,
+        "STEPS",
+    )
     .optopt(
         NORMALISATION_METHOD_SHORT,
         NORMALISATION_METHOD,
@@ -801,7 +814,9 @@ fn get_setup() -> Setup {
         ALSA_MIXER_CONTROL,
     ] {
         if opt_present(a) {
-            warn!("Alsa specific options have no effect if the alsa backend is not enabled at build time.");
+            warn!(
+                "Alsa specific options have no effect if the alsa backend is not enabled at build time."
+            );
             break;
         }
     }
@@ -1105,7 +1120,7 @@ fn get_setup() -> Setup {
     let tmp_dir = opt_str(TEMP_DIR).map_or(SessionConfig::default().tmp_dir, |p| {
         let tmp_dir = PathBuf::from(p);
         if let Err(e) = create_dir_all(&tmp_dir) {
-            error!("could not create or access specified tmp directory: {}", e);
+            error!("could not create or access specified tmp directory: {e}");
             exit(1);
         }
         tmp_dir
@@ -1155,15 +1170,14 @@ fn get_setup() -> Setup {
 
         if audio_dir.is_none() && opt_present(CACHE_SIZE_LIMIT) {
             warn!(
-                "Without a `--{}` / `-{}` path, and/or if the `--{}` / `-{}` flag is set, `--{}` / `-{}` has no effect.",
-                CACHE, CACHE_SHORT, DISABLE_AUDIO_CACHE, DISABLE_AUDIO_CACHE_SHORT, CACHE_SIZE_LIMIT, CACHE_SIZE_LIMIT_SHORT
+                "Without a `--{CACHE}` / `-{CACHE_SHORT}` path, and/or if the `--{DISABLE_AUDIO_CACHE}` / `-{DISABLE_AUDIO_CACHE_SHORT}` flag is set, `--{CACHE_SIZE_LIMIT}` / `-{CACHE_SIZE_LIMIT_SHORT}` has no effect."
             );
         }
 
         let cache = match Cache::new(cred_dir.clone(), volume_dir, audio_dir, limit) {
             Ok(cache) => Some(cache),
             Err(e) => {
-                warn!("Cannot create cache: {}", e);
+                warn!("Cannot create cache: {e}");
                 None
             }
         };
@@ -1188,7 +1202,9 @@ fn get_setup() -> Setup {
                 empty_string_error_msg(USERNAME, USERNAME_SHORT);
             }
             if opt_present(PASSWORD) {
-                error!("Invalid `--{PASSWORD}` / `-{PASSWORD_SHORT}`: Password authentication no longer supported, use OAuth");
+                error!(
+                    "Invalid `--{PASSWORD}` / `-{PASSWORD_SHORT}`: Password authentication no longer supported, use OAuth"
+                );
                 exit(1);
             }
             match cached_creds {
@@ -1217,8 +1233,7 @@ fn get_setup() -> Setup {
         Some("librespot compiled without zeroconf backend".to_owned())
     } else if opt_present(DISABLE_DISCOVERY) {
         Some(format!(
-            "the `--{}` / `-{}` flag set",
-            DISABLE_DISCOVERY, DISABLE_DISCOVERY_SHORT,
+            "the `--{DISABLE_DISCOVERY}` / `-{DISABLE_DISCOVERY_SHORT}` flag set",
         ))
     } else {
         None
@@ -1232,8 +1247,7 @@ fn get_setup() -> Setup {
     let oauth_port = if opt_present(OAUTH_PORT) {
         if !enable_oauth {
             warn!(
-                "Without the `--{}` / `-{}` flag set `--{}` / `-{}` has no effect.",
-                ENABLE_OAUTH, ENABLE_OAUTH_SHORT, OAUTH_PORT, OAUTH_PORT_SHORT
+                "Without the `--{ENABLE_OAUTH}` / `-{ENABLE_OAUTH_SHORT}` flag set `--{OAUTH_PORT}` / `-{OAUTH_PORT_SHORT}` has no effect."
             );
         }
         opt_str(OAUTH_PORT)
@@ -1259,10 +1273,7 @@ fn get_setup() -> Setup {
 
     if let Some(reason) = no_discovery_reason.as_deref() {
         if opt_present(ZEROCONF_PORT) {
-            warn!(
-                "With {} `--{}` / `-{}` has no effect.",
-                reason, ZEROCONF_PORT, ZEROCONF_PORT_SHORT
-            );
+            warn!("With {reason} `--{ZEROCONF_PORT}` / `-{ZEROCONF_PORT_SHORT}` has no effect.");
         }
     }
 
@@ -1340,8 +1351,7 @@ fn get_setup() -> Setup {
     if let Some(reason) = no_discovery_reason.as_deref() {
         if opt_present(ZEROCONF_BACKEND) {
             warn!(
-                "With {} `--{}` / `-{}` has no effect.",
-                reason, ZEROCONF_BACKEND, ZEROCONF_BACKEND_SHORT
+                "With {reason} `--{ZEROCONF_BACKEND}` / `-{ZEROCONF_BACKEND_SHORT}` has no effect."
             );
         }
     }
@@ -1389,31 +1399,31 @@ fn get_setup() -> Setup {
                     name.clone()
                 };
 
-                env::set_var("PULSE_PROP_application.name", pulseaudio_name);
+                set_env_var("PULSE_PROP_application.name", pulseaudio_name).await;
             }
 
             if env::var("PULSE_PROP_application.version").is_err() {
-                env::set_var("PULSE_PROP_application.version", version::SEMVER);
+                set_env_var("PULSE_PROP_application.version", version::SEMVER).await;
             }
 
             if env::var("PULSE_PROP_application.icon_name").is_err() {
-                env::set_var("PULSE_PROP_application.icon_name", "audio-x-generic");
+                set_env_var("PULSE_PROP_application.icon_name", "audio-x-generic").await;
             }
 
             if env::var("PULSE_PROP_application.process.binary").is_err() {
-                env::set_var("PULSE_PROP_application.process.binary", "librespot");
+                set_env_var("PULSE_PROP_application.process.binary", "librespot").await;
             }
 
             if env::var("PULSE_PROP_stream.description").is_err() {
-                env::set_var("PULSE_PROP_stream.description", "Spotify Connect endpoint");
+                set_env_var("PULSE_PROP_stream.description", "Spotify Connect endpoint").await;
             }
 
             if env::var("PULSE_PROP_media.software").is_err() {
-                env::set_var("PULSE_PROP_media.software", "Spotify");
+                set_env_var("PULSE_PROP_media.software", "Spotify").await;
             }
 
             if env::var("PULSE_PROP_media.role").is_err() {
-                env::set_var("PULSE_PROP_media.role", "music");
+                set_env_var("PULSE_PROP_media.role", "music").await;
             }
         }
 
@@ -1431,14 +1441,11 @@ fn get_setup() -> Setup {
                         #[cfg(feature = "alsa-backend")]
                         let default_value = &format!(
                             "{}, or the current value when the alsa mixer is used.",
-                            connect_default_config.initial_volume.unwrap_or_default()
+                            connect_default_config.initial_volume
                         );
 
                         #[cfg(not(feature = "alsa-backend"))]
-                        let default_value = &connect_default_config
-                            .initial_volume
-                            .unwrap_or_default()
-                            .to_string();
+                        let default_value = &connect_default_config.initial_volume.to_string();
 
                         invalid_error_msg(
                             INITIAL_VOLUME,
@@ -1460,7 +1467,8 @@ fn get_setup() -> Setup {
                 } else {
                     cache.as_ref().and_then(Cache::volume)
                 }
-            });
+            })
+            .unwrap_or_default();
 
         let device_type = opt_str(DEVICE_TYPE)
             .as_deref()
@@ -1474,7 +1482,7 @@ fn get_setup() -> Setup {
                         speaker, tv, avr, stb, audiodongle, \
                         gameconsole, castaudio, castvideo, \
                         automobile, smartwatch, chromebook, \
-                        carthing, homething",
+                        carthing",
                         DeviceType::default().into(),
                     );
 
@@ -1483,16 +1491,34 @@ fn get_setup() -> Setup {
             })
             .unwrap_or_default();
 
-        let is_group = opt_present(DEVICE_IS_GROUP);
+        let volume_steps = opt_str(VOLUME_STEPS)
+            .map(|steps| match steps.parse::<u16>() {
+                Ok(value) => value,
+                _ => {
+                    let default_value = &connect_default_config.volume_steps.to_string();
 
-        let has_volume_ctrl = !matches!(mixer_config.volume_ctrl, VolumeCtrl::Fixed);
+                    invalid_error_msg(
+                        VOLUME_STEPS,
+                        VOLUME_STEPS_SHORT,
+                        &steps,
+                        "a positive whole number <= 65535",
+                        default_value,
+                    );
+
+                    exit(1);
+                }
+            })
+            .unwrap_or_else(|| connect_default_config.volume_steps);
+
+        let is_group = opt_present(DEVICE_IS_GROUP);
 
         ConnectConfig {
             name,
             device_type,
             is_group,
             initial_volume,
-            has_volume_ctrl,
+            volume_steps,
+            ..connect_default_config
         }
     };
 
@@ -1510,7 +1536,7 @@ fn get_setup() -> Setup {
                         url
                     },
                     Err(e) => {
-                        error!("Invalid proxy URL: \"{}\", only URLs in the format \"http(s)://host:port\" are allowed", e);
+                        error!("Invalid proxy URL: \"{e}\", only URLs in the format \"http(s)://host:port\" are allowed");
                         exit(1);
                     }
                 }
@@ -1567,8 +1593,7 @@ fn get_setup() -> Setup {
             ] {
                 if opt_present(a) {
                     warn!(
-                        "Without the `--{}` / `-{}` flag normalisation options have no effect.",
-                        ENABLE_VOLUME_NORMALISATION, ENABLE_VOLUME_NORMALISATION_SHORT,
+                        "Without the `--{ENABLE_VOLUME_NORMALISATION}` / `-{ENABLE_VOLUME_NORMALISATION_SHORT}` flag normalisation options have no effect.",
                     );
                     break;
                 }
@@ -1750,7 +1775,7 @@ fn get_setup() -> Setup {
                 "none" => None,
                 _ => match format {
                     AudioFormat::F64 | AudioFormat::F32 => {
-                        error!("Dithering is not available with format: {:?}.", format);
+                        error!("Dithering is not available with format: {format:?}.");
                         exit(1);
                     }
                     _ => Some(dither::find_ditherer(ditherer_name).unwrap_or_else(|| {
@@ -1792,6 +1817,7 @@ fn get_setup() -> Setup {
             normalisation_release_cf,
             normalisation_knee_db,
             ditherer,
+            position_update_interval: None,
         }
     };
 
@@ -1819,6 +1845,23 @@ fn get_setup() -> Setup {
     }
 }
 
+// Initialize a static semaphore with only one permit, which is used to
+// prevent setting environment variables from running in parallel.
+static PERMIT: Semaphore = Semaphore::const_new(1);
+async fn set_env_var<K: AsRef<OsStr>, V: AsRef<OsStr>>(key: K, value: V) {
+    let permit = PERMIT
+        .acquire()
+        .await
+        .expect("Failed to acquire semaphore permit");
+
+    // SAFETY: This is safe because setting the environment variable will wait if the permit is
+    // already acquired by other callers.
+    unsafe { env::set_var(key, value) }
+
+    // Drop the permit manually, so the compiler doesn't optimize it away as unused variable.
+    drop(permit);
+}
+
 #[tokio::main(flavor = "current_thread")]
 async fn main() {
     const RUST_BACKTRACE: &str = "RUST_BACKTRACE";
@@ -1827,10 +1870,10 @@ async fn main() {
     const RECONNECT_RATE_LIMIT: usize = 5;
 
     if env::var(RUST_BACKTRACE).is_err() {
-        env::set_var(RUST_BACKTRACE, "full")
+        set_env_var(RUST_BACKTRACE, "full").await;
     }
 
-    let setup = get_setup();
+    let setup = get_setup().await;
 
     let mut last_credentials = None;
     let mut spirc: Option<Spirc> = None;
@@ -1869,7 +1912,7 @@ async fn main() {
             {
                 Ok(d) => break Some(d),
                 Err(e) => {
-                    sys.refresh_processes(ProcessesToUpdate::All);
+                    sys.refresh_processes(ProcessesToUpdate::All, true);
 
                     if System::uptime() <= 1 {
                         debug!("Retrying to initialise discovery: {e}");
@@ -1892,18 +1935,22 @@ async fn main() {
             Some(port) => format!(":{port}"),
             _ => String::new(),
         };
-        let access_token = match librespot::oauth::get_access_token(
+        let client = OAuthClientBuilder::new(
             &setup.session_config.client_id,
             &format!("http://127.0.0.1{port_str}/login"),
             OAUTH_SCOPES.to_vec(),
-        ) {
-            Ok(token) => token.access_token,
-            Err(e) => {
-                error!("Failed to get Spotify access token: {e}");
-                exit(1);
-            }
-        };
-        last_credentials = Some(Credentials::with_access_token(access_token));
+        )
+        .open_in_browser()
+        .build()
+        .unwrap_or_else(|e| {
+            error!("Failed to create OAuth client: {e}");
+            exit(1);
+        });
+        let oauth_token = client.get_access_token().unwrap_or_else(|e| {
+            error!("Failed to get Spotify access token: {e}");
+            exit(1);
+        });
+        last_credentials = Some(Credentials::with_access_token(oauth_token.access_token));
         connecting = true;
     } else if discovery.is_none() {
         error!(
@@ -1913,7 +1960,13 @@ async fn main() {
     }
 
     let mixer_config = setup.mixer_config.clone();
-    let mixer = (setup.mixer)(mixer_config);
+    let mixer = match (setup.mixer)(mixer_config) {
+        Ok(mixer) => mixer,
+        Err(why) => {
+            error!("{why}");
+            exit(1)
+        }
+    };
     let player_config = setup.player_config.clone();
 
     let soft_volume = mixer.get_soft_volume();
@@ -1952,7 +2005,7 @@ async fn main() {
 
                         if let Some(spirc) = spirc.take() {
                             if let Err(e) = spirc.shutdown() {
-                                error!("error sending spirc shutdown message: {}", e);
+                                error!("error sending spirc shutdown message: {e}");
                             }
                         }
                         if let Some(spirc_task) = spirc_task.take() {
@@ -1986,7 +2039,7 @@ async fn main() {
                                                                 mixer.clone()).await {
                     Ok((spirc_, spirc_task_)) => (spirc_, spirc_task_),
                     Err(e) => {
-                        error!("could not initialize spirc: {}", e);
+                        error!("could not initialize spirc: {e}");
                         exit(1);
                     }
                 };
@@ -2038,7 +2091,7 @@ async fn main() {
     // Shutdown spirc if necessary
     if let Some(spirc) = spirc {
         if let Err(e) = spirc.shutdown() {
-            error!("error sending spirc shutdown message: {}", e);
+            error!("error sending spirc shutdown message: {e}");
         }
 
         if let Some(spirc_task) = spirc_task {
